/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"
#include <vector>

using namespace std;
using namespace FD;

class CvfdhHoughLine;
   
DECLARE_NODE(CvfdhHoughLine)
/*Node
 *
 * @name CvfdhHoughLine
 * @category FDOPENCV_HARK:Segmentation
 * @description CvfdhHoughLine
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @parameter_name ALGORITHM
 * @parameter_type string
 * @parameter_value STANDARD
 * @parameter_list STANDARD:PROBABLISTIC
 * @parameter_description
 *
 END*/
   
   
class CvfdhHoughLine : public BufferedNode {

  int m_imageInID;
  int m_srcOutID;
  string algorithm;
      
public:
  CvfdhHoughLine(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcOutID = addOutput("SRCOUT");
    algorithm = object_cast<String>(parameters.get("ALGORITHM"));
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    Vector<ObjectRef>& sources = *new Vector<ObjectRef>;
    (*(outputs[m_srcOutID].buffer))[count] = &sources;

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){
      IplImage* img_in;      
      img_in = cvCloneImage((IplImage*)image_vec[cnt]);
    
      // ---------------------------------------------------------------
      // main routine : start  

      if(img_in->nChannels > 1){
	IplImage *gray;
	gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
	cvCvtColor(img_in, gray, CV_BGR2GRAY);
	img_in = cvCloneImage (gray);
	cvReleaseImage (&gray);
      }    

      IplImage* dst = cvCreateImage( cvGetSize(img_in), 8, 1 );
      IplImage* color_dst = cvCreateImage( cvGetSize(img_in), 8, 3 );
      CvMemStorage* storage = cvCreateMemStorage(0);
      CvSeq* lines = 0;
      int i;
      cvCanny( img_in, dst, 50, 200, 3);
      cvCvtColor( dst, color_dst, CV_GRAY2BGR );

      if(algorithm == "STANDARD"){
	lines = cvHoughLines2( dst, storage, CV_HOUGH_STANDARD, 1, CV_PI/180, 100, 0, 0 );
	for( i = 0; i < MIN(lines->total,100); i++ ){
	  float* line = (float*)cvGetSeqElem(lines,i);
	  float rho = line[0];
	  float theta = line[1];
	  CvPoint pt1, pt2;
	  double a = cos(theta), b = sin(theta);
	  double x0 = a*rho, y0 = b*rho;
	  pt1.x = cvRound(x0 + 1000*(-b));
	  pt1.y = cvRound(y0 + 1000*(a));
	  pt2.x = cvRound(x0 - 1000*(-b));
	  pt2.y = cvRound(y0 - 1000*(a));
	  // cvLine( color_dst, pt1, pt2, CV_RGB(255,0,0), 3, 8 );

	  RCPtr<SrcInfo> src(new SrcInfo);
	  src->roi_c[0] = (int)((pt1.x + pt2.x) / 2);
	  src->roi_c[1] = (int)((pt1.y + pt2.y) / 2);
	  src->roi_xy[0] = abs((int)(pt1.x - pt2.x));
	  src->roi_xy[1] = abs((int)(pt1.y - pt2.y));
	  src->img_size[0] = img_in->width;
	  src->img_size[1] = img_in->height;

	  sources.push_back(src);

	}
      }

      if(algorithm == "PROBABLISTIC"){
	lines = cvHoughLines2( dst, storage, CV_HOUGH_PROBABILISTIC, 1, CV_PI/180, 50, 50, 10 );
	for( i = 0; i < lines->total; i++ ){
	  CvPoint* line = (CvPoint*)cvGetSeqElem(lines,i);
	  // cvLine( color_dst, line[0], line[1], CV_RGB(255,0,0), 3, 8 );

	  RCPtr<SrcInfo> src(new SrcInfo);
	  src->roi_c[0] = (int)((line[0].x + line[1].x) / 2);
	  src->roi_c[1] = (int)((line[0].y + line[1].y) / 2);
	  src->roi_xy[0] = abs((int)(line[0].x - line[1].x));
	  src->roi_xy[1] = abs((int)(line[0].y - line[1].y));
	  src->img_size[0] = img_in->width;
	  src->img_size[1] = img_in->height;

	  sources.push_back(src);

	} 
      }

      cvReleaseImage (&img_in);
      cvReleaseImage (&dst);
      cvReleaseImage (&color_dst);

    }

    cvReleaseImage(&img_in_tmp);

  }
   
  NO_ORDER_NODE_SPEEDUP(CvfdhHoughLine)
};
